@@ -7,6 +7,7 @@ import flammkuchen
 import numpy as np
 from . import utils, data, models, event_utils, segment_utils, annot
 from typing import List, Optional, Dict
+import glob
 
 from tensorflow.python.framework.ops import disable_eager_execution
 disable_eager_execution()
@@ -86,13 +87,15 @@ def predict_segments(class_probabilities: np.array,
         segment_dims (Optional[List[int]], optional): set of indices into class_probabilities corresponding
                                                       to segment-like song types.
                                                       Needs to include the noise dim.
-                                                      Defaults to None.
-        segment_names (Optional[List[str]], optional): [description]. Defaults to None.
+                                                      Required to ignore event-like song types.
+                                                      Defaults to None (all classes are considered segment-like).
+        segment_names (Optional[List[str]], optional): Names for segment-like classes.
+                                                       Defaults to None (use indices of segment-like classes).
         segment_ref_onsets (Optional[List[float]], optional):
-                            Use onsets (in seconds) as ref for estimating labels.
+                            Syllable onsets (in seconds) to use for estimating labels.
                             Defaults to None (will use onsets est from class_probabilitieslabels as ref).
         segment_ref_offsets (Optional[List[float]], optional): [description].
-                            Use offsets (in seconds) as ref for estimating labels.
+                            Syllable offsets (in seconds) to use for estimating labels.
                             Defaults to None (will use offsets est from class_probabilitieslabels as ref).
         segment_thres (float, optional): [description]. Defaults to 0.5.
         segment_minlen (Optional[float], optional): seconds. Defaults to None.
@@ -253,8 +256,8 @@ def predict(x: np.array, model_save_name: str = None, verbose: int = 1,
             model: models.keras.models.Model = None, params: dict = None,
             event_thres: float = 0.5, event_dist: float = 0.01,
             event_dist_min: float = 0, event_dist_max: float = None,
-            segment_thres: float = 0.5, segment_minlen: float = None,
-            segment_fillgap: float = None,
+            segment_thres: float = 0.5, segment_use_optimized: bool = True,
+            segment_minlen: float = None, segment_fillgap: float = None,
             pad: bool = True, prepend_data_padding: bool = True):
     """[summary]
 
@@ -293,8 +296,13 @@ def predict(x: np.array, model_save_name: str = None, verbose: int = 1,
                                 Defaults to None (no upper limit).
 
         segment_thres (float): Confidence threshold for detecting segments. Range 0..1. Defaults to 0.5.
+        segment_use_optimized (bool): Use minlen and fillgap values from param file if they exist.
+                                      If segment_minlen and segment_fillgap are provided,
+                                      then they will override the values from the param file.
+                                      Defaults to True.
         segment_minlen (float): Minimal duration in seconds of a segment used for filtering out spurious detections. Defaults to None.
         segment_fillgap (float): Gap in seconds between adjacent segments to be filled. Useful for correcting brief lapses. Defaults to None.
+
         pad (bool): prepend values (repeat last sample value) to fill the last batch. Otherwise, the end of the data will not be annotated because
                     the last, non-full batch will be skipped.
         prepend_data_padding (bool, optional): Restores samples that are ignored
@@ -316,6 +324,13 @@ def predict(x: np.array, model_save_name: str = None, verbose: int = 1,
         assert isinstance(model, models.keras.models.Model)
         assert isinstance(params, dict)
 
+    # use postprocessing values from params and/or args
+    if segment_use_optimized and 'post_opt' in params and isinstance(params['post_opt'], dict):
+        if segment_minlen is None:
+            segment_minlen = params['post_opt']['min_len']
+        if segment_fillgap is None:
+            segment_fillgap = params['post_opt']['gap_dur']
+
     if batch_size is not None:
         params['batch_size'] = batch_size
 
@@ -342,27 +357,33 @@ def predict(x: np.array, model_save_name: str = None, verbose: int = 1,
     return events, segments, class_probabilities, params['class_names']
 
 
-def cli_predict(recording_filename: str, model_save_name: str, *,
-                save_filename: Optional[str] = None, save_format: str = 'h5',
+def cli_predict(path: str, model_save_name: str, *,
+                save_filename: Optional[str] = None, save_format: str = 'csv',
                 verbose: int = 1, batch_size: Optional[int] = None,
                 event_thres: float = 0.5, event_dist: float = 0.01,
                 event_dist_min: float = 0, event_dist_max: Optional[float] = None,
-                segment_thres: float = 0.5, segment_minlen: Optional[float] = None,
-                segment_fillgap: Optional[float] = None):
-    """Predict song labels in a wav file.
+                segment_thres: float = 0.5, segment_use_optimized: Optional[bool] = None,
+                segment_minlen: Optional[float] = None, segment_fillgap: Optional[float] = None):
+    """Predict song labels for a wav file or a folder of wav files.
 
-    Saves hdf5 file with keys: events, segments, class_probabilities
+    Saves hdf5 files with keys: events, segments, class_probabilities
+    OR csv files with columns: label/start_seconds/stop_seconds
 
     Args:
-        recording_filename (str): Path to the WAV file with the audio data.
+        path (str): Path to a single WAV file with the audio data or to a folder with WAV files.
         model_save_name (str): Stem of the path for the model (and parameters). File to load will be MODEL_SAVE_NAME + _model.h5.
         save_filename (Optional[str]): Path to save annotations to.
                                        If omitted, will construct save_filename by
                                        stripping the extension from recording_filename and adding '_das.h5' or '_annotations.csv'.
-        save_format (str): 'csv' or 'h5'. Defaults to 'h5'.
+                                       Will be ignored if `path` is a folder.
+        save_format (str): 'csv' or 'h5'.
+                           csv: tabular text file with label, start and end seconds for each predicted song.
+                           h5: same information as in csv plus confidence values for each sample and song type.
+                           Defaults to 'csv'.
         verbose (int): Display progress bar during prediction. Defaults to 1.
         batch_size (Optional[int]): Number of chunks processed at once.
                                     Defaults to None (the default used during training).
+
         event_thres (float): Confidence threshold for detecting events. Range 0..1. Defaults to 0.5.
         event_dist (float): Minimal distance between adjacent events during thresholding.
                             Prevents detecting duplicate events when the confidence trace is a little noisy.
@@ -371,7 +392,12 @@ def cli_predict(recording_filename: str, model_save_name: str, *,
                                 Defaults to 0.
         event_dist_max (Optional[float]): MAXimal inter-event interval for the event filter run during post processing.
                                           Defaults to None (no upper limit).
+
         segment_thres (float): Confidence threshold for detecting segments. Range 0..1. Defaults to 0.5.
+        segment_use_optimized (Optional[bool]): Use minlen and fillgap values from param file if they exist.
+                                      If segment_minlen and segment_fillgap are provided,
+                                      then they will override the values from the param file.
+                                      Defaults to True.
         segment_minlen (Optional[float]): Minimal duration of a segment used for filtering out spurious detections.
                                           Defaults to None (keep all segments).
         segment_fillgap (Optional[float]): Gap between adjacent segments to be filled. Useful for correcting brief lapses.
@@ -383,33 +409,54 @@ def cli_predict(recording_filename: str, model_save_name: str, *,
     if not (save_format == 'csv' or save_format == 'h5'):
         raise ValueError(f"Unknown save_format '{save_format}'. Should be either 'csv' or 'h5'.")
 
-    logging.info(f"   Loading data from {recording_filename}.")
-    _, x = scipy.io.wavfile.read(recording_filename)
-    x = np.atleast_2d(x).T
+    if os.path.isdir(path) and save_filename is not None:
+        logging.warning(f'{path} is a folder. Will ignore save_filename argument {save_filename}.')
 
-    logging.info(f"   Annotating using model at {model_save_name}.")
-    events, segments, class_probabilities, class_names = predict(x, model_save_name, verbose, batch_size,
-                                                    None, None,
-                                                    event_thres, event_dist, event_dist_min, event_dist_max,
-                                                    segment_thres, segment_minlen, segment_fillgap)
+    # if path is folder: glob contents - all files
+    if os.path.isdir(path):
+        filenames = glob.glob(f'{path}/*.wav')
+        filenames = [filename for filename in filenames if not os.path.isdir(filename)]
+    elif os.path.isfile(path):
+        filenames = [path]
 
-    if save_format == 'h5':
-        d = {'events': events,
-            'segments': segments,
-            'class_probabilities': class_probabilities,
-            'class_names': class_names}
+    model, params = utils.load_model_and_params(model_save_name)
 
-        if save_filename is None:
-            save_filename = os.path.splitext(recording_filename)[0] + '_das.h5'
+    for recording_filename in filenames:
+        logging.info(f"   Loading data from {recording_filename}.")
+        try:
+            # else if path is file - predict only on file but make it single-item list
+            _, x = scipy.io.wavfile.read(recording_filename)
+            x = np.atleast_2d(x).T
 
-        logging.info(f"   Saving results to {save_filename}.")
-        flammkuchen.save(save_filename, d)
-        logging.info(f"Done.")
+            logging.info(f"   Annotating using model at {model_save_name}.")
+            # TODO: load model once, provide as direct arg
+            events, segments, class_probabilities, class_names = predict(x, None, verbose, batch_size,
+                                                            model, params,
+                                                            event_thres, event_dist, event_dist_min, event_dist_max,
+                                                            segment_thres, segment_use_optimized, segment_minlen, segment_fillgap)
+
+            if save_format == 'h5':
+                # turn events and segments into df!
+                d = {'events': events,
+                    'segments': segments,
+                    'class_probabilities': class_probabilities,
+                    'class_names': class_names}
+                if save_filename is None:
+                    save_filename = os.path.splitext(recording_filename)[0] + '_das.h5'
+                logging.info(f"   Saving results to {save_filename}.")
+                flammkuchen.save(save_filename, d)
+                logging.info(f"Done.")
+            elif save_format == 'csv':
+                evt = annot.Events.from_predict(events, segments)
+                if save_filename is None:
+                    save_filename = os.path.splitext(recording_filename)[0] + '_annotations.csv'
+                logging.info(f"   Saving results to {save_filename}.")
+                evt.to_df().to_csv(save_filename)
+                logging.info(f"Done.")
+
+            # reset
+            if os.path.isdir(path):
+                save_filename = None
+        except Exception as e:
+            logging.exception(f'Error processing file {recording_filename}.')
 
-    elif save_format == 'csv':
-        df = annot.Events.from_predict(events, segments)
-        if save_filename is None:
-            save_filename = os.path.splitext(recording_filename)[0] + '_annotations.csv'
-        logging.info(f"   Saving results to {save_filename}.")
-        df.to_csv(save_filename)
-        logging.info(f"Done.")
